from logging import getLogger
from typing import Optional
from injector import singleton

from gumo.core.injector import injector
from gumo.core.infrastructure.configuration import GumoConfiguration

logger = getLogger('gumo.core')


def configure(
        google_cloud_project: Optional[str] = None,
        google_cloud_location: Optional[str] = None,
        service_account_credential_path: Optional[str] = None,
):
    if google_cloud_location is not None:
        logger.warning(f'The argument "google_cloud_location" of gumo.core.configure() is deprecated.')
    if service_account_credential_path is not None:
        logger.warning(f'The argument "service_account_credential_path" of gumo.core.configure() is deprecated.')

    config = GumoConfiguration(
        google_cloud_project=google_cloud_project,
    )
    logger.debug(f'Gumo is configured, config={config}')
    injector.binder.bind(GumoConfiguration, to=config, scope=singleton)

    return config


def get_config() -> GumoConfiguration:
    return injector.get(GumoConfiguration)



import os
from typing import Dict

import yaml

from logging import getLogger
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import SafeLoader as Loader

logger = getLogger(__name__)


class MockAppEngineEnvironment:
    @classmethod
    def load_app_yaml(cls, app_yaml_path):
        with open(app_yaml_path, 'r+') as f:
            app_yaml = yaml.load(f, Loader=Loader)

        if 'env_variables' not in app_yaml:
            logger.debug('"env_variables" doest not exists in app.yaml')
        else:
            logger.debug('"env_variables" exists in app.yaml:')
            cls._set_env_variables(env_variables=app_yaml["env_variables"])

    @classmethod
    def _set_env_variables(cls, env_variables: Dict[str, str]) -> None:
        for key, value in env_variables.items():
            if os.environ.get(key):
                logger.debug(f' * os.environ["{key}"] ... skipped (current value= {os.environ.get(key)}')
            else:
                os.environ[key] = str(value)
                logger.debug(f' * os.environ["{key}"] = "{str(value)}"')



import os
import dataclasses
import threading

from typing import Optional
from typing import Union
from typing import ClassVar

from gumo.core.domain.configuration import GoogleCloudProjectID
from gumo.core.domain.configuration import ApplicationPlatform


@dataclasses.dataclass(frozen=False)
class GumoConfiguration:
    google_cloud_project: Union[GoogleCloudProjectID, str, None] = None
    application_platform: Optional[ApplicationPlatform] = None

    _ENV_KEY_GOOGLE_CLOUD_PROJECT: ClassVar = 'GOOGLE_CLOUD_PROJECT'
    _lock: ClassVar = threading.Lock()

    def __post_init__(self):
        with self._lock:
            self._set_google_cloud_project()
            self._set_application_platform()

    def _set_google_cloud_project(self):
        if isinstance(self.google_cloud_project, str):
            self.google_cloud_project = GoogleCloudProjectID(self.google_cloud_project)
        if isinstance(self.google_cloud_project, GoogleCloudProjectID):
            if self.google_cloud_project.value != os.environ.get(self._ENV_KEY_GOOGLE_CLOUD_PROJECT):
                raise RuntimeError(f'Env-var "{self._ENV_KEY_GOOGLE_CLOUD_PROJECT}" is invalid or undefined.'
                                   f'Please set value "{self.google_cloud_project.value}" to env-vars.')

        if self.google_cloud_project is None:
            if self._ENV_KEY_GOOGLE_CLOUD_PROJECT in os.environ:
                self.google_cloud_project = GoogleCloudProjectID(os.environ[self._ENV_KEY_GOOGLE_CLOUD_PROJECT])
            else:
                raise RuntimeError(f'Env-var "{self._ENV_KEY_GOOGLE_CLOUD_PROJECT}" is undefined, please set it.')

    def _set_application_platform(self):
        is_google_app_engine = 'GAE_DEPLOYMENT_ID' in os.environ and 'GAE_INSTANCE' in os.environ
        is_google_compute_engine = 'COMPUTE_ENGINE' in os.environ
        if is_google_app_engine:
            self.application_platform = ApplicationPlatform.GoogleAppEngine
        elif is_google_compute_engine:
            self.application_platform = ApplicationPlatform.GoogleComputeEngine
        else:
            self.application_platform = ApplicationPlatform.Local

    @property
    def is_local(self) -> bool:
        return self.application_platform == ApplicationPlatform.Local

    @property
    def is_google_app_engine(self) -> bool:
        return self.application_platform == ApplicationPlatform.GoogleAppEngine

    @property
    def is_google_compute_engine(self) -> bool:
        return self.application_platform == ApplicationPlatform.GoogleComputeEngine

    @property
    def is_google_platform(self) -> bool:
        return self.is_google_app_engine or self.is_google_compute_engine



from gumo.core._configuration import configure
from gumo.core._configuration import get_config

from gumo.core.exceptions import ConfigurationError

from gumo.core.infrastructure.configuration import GumoConfiguration
from gumo.core.domain.configuration import GoogleCloudProjectID

from gumo.core.domain.entity_key import EntityKey
from gumo.core.domain.entity_key import NoneKey
from gumo.core.domain.entity_key import EntityKeyFactory

from gumo.core.application.entity_key import EntityKeyGenerator

from gumo.core.infrastructure import MockAppEngineEnvironment
from gumo.core.infrastructure.credential import get_google_oauth_credentials
from gumo.core.infrastructure.credential import get_google_id_token_credentials


# backward compatibility
get_google_oauth_credential = get_google_oauth_credentials
get_google_id_token_credential = get_google_id_token_credentials


__all__ = [
    configure.__name__,
    get_config.__name__,

    ConfigurationError.__name__,

    GumoConfiguration.__name__,
    GoogleCloudProjectID.__name__,

    EntityKey.__name__,
    NoneKey.__name__,
    EntityKeyFactory.__name__,

    EntityKeyGenerator.__name__,

    MockAppEngineEnvironment.__name__,
    get_google_oauth_credentials.__name__,
    get_google_id_token_credentials.__name__,

    # backward compatibility:
    get_google_oauth_credential.__name__,
    get_google_id_token_credential.__name__,
]


configure()



import dataclasses
import enum


@dataclasses.dataclass(frozen=True)
class GoogleCloudProjectID:
    value: str

    def __post_init__(self):
        if not isinstance(self.value, str):
            raise ValueError(f'project_id must be a string, expect: {type(self.value)}')


class ApplicationPlatform(enum.Enum):
    Local = 'local'
    GoogleAppEngine = 'google-app-engine'
    GoogleComputeEngine = 'google-compute-engine'



import os
import json

from logging import getLogger
from injector import singleton

from typing import Optional
from typing import Union
from typing import Tuple

import google.auth.transport.requests
from google.oauth2 import service_account
from google.auth import compute_engine
from injector import inject

from gumo.core.exceptions import ServiceAccountConfigurationError
from gumo.core.injector import injector
from gumo.core.infrastructure.configuration import GumoConfiguration

logger = getLogger(__name__)

DEFAULT_TOKEN_URI = 'https://www.googleapis.com/oauth2/v4/token'

Credentials = Union[service_account.Credentials, compute_engine.Credentials]
IDTokenCredentials = Union[service_account.IDTokenCredentials, compute_engine.IDTokenCredentials]
Request = google.auth.transport.Request


class _GoogleCredentialManagerForServiceAccountCredential:
    def __init__(
            self,
            credential_file_path: str,
    ):
        self._credential_file_path = credential_file_path

        self._credentials: Optional[Credentials] = None
        self._cache_id_token_credentials = {}

    def build_credentials(self) -> Credentials:
        try:
            if not self._credentials:
                logger.debug('Loading service_account.Credentials from local file.')
                self._credentials = service_account.Credentials.from_service_account_info(
                    info=self._get_content_from_local_file()
                )

            return self._credentials
        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

    def _build_request(self) -> Request:
        return google.auth.transport.requests.Request()

    def _fetch_id_token_credentials(self, target_audience: str) -> Optional[IDTokenCredentials]:
        return self._cache_id_token_credentials.get(target_audience)

    def build_id_token_credentials(
            self,
            target_audience: str,
            token_uri: Optional[str] = None,
    ) -> Tuple[IDTokenCredentials, Request]:
        request = self._build_request()
        id_token_credentials = self._fetch_id_token_credentials(target_audience=target_audience)

        try:
            if not id_token_credentials or not id_token_credentials.valid:
                logger.debug('Build service_account.IDTokenCredentials and token refresh.')

                _credential = self.build_credentials()
                id_token_credentials = service_account.IDTokenCredentials(
                    signer=_credential.signer,
                    service_account_email=_credential.service_account_email,
                    token_uri=token_uri if token_uri else DEFAULT_TOKEN_URI,
                    target_audience=target_audience,
                )
                id_token_credentials.refresh(request=request)

                self._cache_id_token_credentials[target_audience] = id_token_credentials

            return (id_token_credentials, request)
        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

    def _get_content_from_local_file(self):
        if not os.path.exists(self._credential_file_path):
            raise ServiceAccountConfigurationError(
                f'GOOGLE_APPLICATION_CREDENTIALS={self._credential_file_path} is not found.'
            )

        with open(self._credential_file_path, 'r') as f:
            content = f.read()

        return json.loads(content)


class _GoogleCredentialManagerForComputeEngine:
    def __init__(self):
        self._credentials: Optional[Credentials] = None
        self._cache_id_token_credentials = {}

    def _build_request(self) -> Request:
        return google.auth.transport.requests.Request()

    def build_credentials(self) -> Credentials:
        try:
            if not self._credentials or not self._credentials.valid:
                logger.debug('Fetch compute_engine.Credentials and token refresh.')

                self._credentials = compute_engine.Credentials()
                self._credentials.refresh(request=self._build_request())

            return self._credentials
        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

    def _fetch_id_token_credentials(self, target_audience: str) -> Optional[IDTokenCredentials]:
        return self._cache_id_token_credentials.get(target_audience)

    def build_id_token_credentials(
            self,
            target_audience: str,
            token_uri: Optional[str] = None,
    ) -> Tuple[IDTokenCredentials, Request]:
        request = self._build_request()
        id_token_credentials = self._fetch_id_token_credentials(target_audience=target_audience)

        try:
            if not id_token_credentials or not id_token_credentials.valid:
                logger.debug('Fetch compute_engine.IDTokenCredentials and token refresh.')

                id_token_credentials = compute_engine.IDTokenCredentials(
                    request=request,
                    target_audience=target_audience,
                    token_uri=token_uri if token_uri else DEFAULT_TOKEN_URI,
                )
                id_token_credentials.refresh(request=request)

                self._cache_id_token_credentials[target_audience] = id_token_credentials

        except ServiceAccountConfigurationError:
            raise
        except RuntimeError as e:
            raise ServiceAccountConfigurationError(e)

        return (id_token_credentials, request)


class GoogleOAuthCredentialManager:
    _ENV_KEY_GOOGLE_APPLICATION_CREDENTIALS = 'GOOGLE_APPLICATION_CREDENTIALS'

    @inject
    def __init__(
            self,
            gumo_configuration: GumoConfiguration,
    ):
        self._gumo_configuration = gumo_configuration
        if self._ENV_KEY_GOOGLE_APPLICATION_CREDENTIALS in os.environ:
            self._manager = _GoogleCredentialManagerForServiceAccountCredential(
                credential_file_path=os.environ.get(self._ENV_KEY_GOOGLE_APPLICATION_CREDENTIALS)
            )
        elif self._gumo_configuration.is_google_platform:
            self._manager = _GoogleCredentialManagerForComputeEngine()
        else:
            raise RuntimeError(f'Unknown Platform configuration of GumoConfiguration.')

    def build_credentials(self) -> Credentials:
        return self._manager.build_credentials()

    def _build_request(self) -> Request:
        return google.auth.transport.requests.Request()

    def build_id_token_credentials(
            self,
            target_audience: str,
            with_refresh: bool = True,
            token_uri: Optional[str] = None,
    ) -> Tuple[IDTokenCredentials, Request]:
        _ = with_refresh  # deprecated arguments.

        return self._manager.build_id_token_credentials(
            target_audience=target_audience,
            token_uri=token_uri,
        )


def get_google_oauth_credentials() -> Credentials:
    return injector.get(GoogleOAuthCredentialManager, scope=singleton).build_credentials()


def get_google_id_token_credentials(
        target_audience: str,
        with_refresh: bool = True,
        token_uri: Optional[str] = None,
) -> Tuple[IDTokenCredentials, Request]:
    return injector.get(GoogleOAuthCredentialManager, scope=singleton).build_id_token_credentials(
        target_audience=target_audience,
        with_refresh=with_refresh,
        token_uri=token_uri,
    )
